@@ -3,12 +3,12 @@ ldap - base module
 
 See http://www.python-ldap.org/ for details.
 
-$Id: __init__.py,v 1.102 2016/01/18 15:17:30 stroeder Exp $
+$Id: __init__.py,v 1.119 2017/07/12 17:30:12 stroeder Exp $
 """
 
 # This is also the overall release version number
 
-__version__ = '2.4.25'
+__version__ = '2.4.41'
 
 import sys
 
@@ -82,7 +82,7 @@ class LDAPLock:
 # Create module-wide lock for serializing all calls into underlying LDAP lib
 _ldap_module_lock = LDAPLock(desc='Module wide')
 
-from functions import open,initialize,init,get_option,set_option,escape_str
+from functions import open,initialize,init,get_option,set_option,escape_str,strf_secs,strp_secs
 
 from ldapobject import NO_UNIQUE_ENTRY
 
