@@ -3,7 +3,7 @@ ldapobject.py - wraps class _ldap.LDAPObject
 
 See http://www.python-ldap.org/ for details.
 
-\$Id: ldapobject.py,v 1.143 2014/11/23 18:51:53 stroeder Exp $
+\$Id: ldapobject.py,v 1.163 2017/05/29 09:00:15 stroeder Exp $
 
 Compability:
 - Tested with Python 2.0+ but should work with Python 1.5.x
@@ -18,6 +18,8 @@ Basically calls into the LDAP lib are serialized by the module-wide
 lock self._ldap_object_lock.
 """
 
+from os import strerror
+
 from ldap import __version__
 
 __all__ = [
@@ -32,7 +34,7 @@ if __debug__:
   # Tracing is only supported in debugging mode
   import traceback
 
-import sys,time,pprint,_ldap,ldap,ldap.functions
+import sys,time,pprint,_ldap,ldap,ldap.sasl,ldap.functions
 
 from ldap.schema import SCHEMA_ATTRS
 from ldap.controls import LDAPControl,DecodeControlTuples,RequestControlTuples
@@ -109,7 +111,12 @@ class SimpleLDAPObject:
             diagnostic_message_success = self._l.get_option(ldap.OPT_DIAGNOSTIC_MESSAGE)
       finally:
         self._ldap_object_lock.release()
-    except LDAPError,e:
+    except LDAPError, e:
+      try:
+        if not e.args[0].has_key('info') and e.args[0].has_key('errno'):
+          e.args[0]['info'] = strerror(e.args[0]['errno'])
+      except IndexError:
+        pass
       if __debug__ and self._trace_level>=2:
         self._trace_file.write('=> LDAPError - %s: %s\n' % (e.__class__.__name__,str(e)))
       raise
@@ -136,6 +143,14 @@ class SimpleLDAPObject:
         self.__class__.__name__,repr(name)
       )
 
+  def fileno(self):
+    """
+    Returns file description of LDAP connection.
+
+    Just a convenience wrapper for LDAPObject.get_option(ldap.OPT_DESC)
+    """
+    return self.get_option(ldap.OPT_DESC)
+
   def abandon_ext(self,msgid,serverctrls=None,clientctrls=None):
     """
     abandon_ext(msgid[,serverctrls=None[,clientctrls=None]]) -> None
@@ -198,8 +213,7 @@ class SimpleLDAPObject:
     return self.add_ext(dn,modlist,None,None)
 
   def add_s(self,dn,modlist):
-    msgid = self.add(dn,modlist)
-    return self.result(msgid,all=1,timeout=self.timeout)
+    return self.add_ext_s(dn,modlist,None,None)
 
   def simple_bind(self,who='',cred='',serverctrls=None,clientctrls=None):
     """
@@ -235,6 +249,36 @@ class SimpleLDAPObject:
     """
     return self._ldap_call(self._l.sasl_interactive_bind_s,who,auth,RequestControlTuples(serverctrls),RequestControlTuples(clientctrls),sasl_flags)
 
+  def sasl_non_interactive_bind_s(self,sasl_mech,serverctrls=None,clientctrls=None,sasl_flags=ldap.SASL_QUIET,authz_id=''):
+    """
+    Send a SASL bind request using a non-interactive SASL method (e.g. GSSAPI, EXTERNAL)
+    """
+    self.sasl_interactive_bind_s(
+      '',
+      ldap.sasl.sasl(
+        {ldap.sasl.CB_USER:authz_id},
+        sasl_mech
+      )
+    )
+
+  def sasl_external_bind_s(self,serverctrls=None,clientctrls=None,sasl_flags=ldap.SASL_QUIET,authz_id=''):
+    """
+    Send SASL bind request using SASL mech EXTERNAL
+    """
+    self.sasl_non_interactive_bind_s('EXTERNAL',serverctrls,clientctrls,sasl_flags,authz_id)
+
+  def sasl_gssapi_bind_s(self,serverctrls=None,clientctrls=None,sasl_flags=ldap.SASL_QUIET,authz_id=''):
+    """
+    Send SASL bind request using SASL mech GSSAPI
+    """
+    self.sasl_non_interactive_bind_s('GSSAPI',serverctrls,clientctrls,sasl_flags,authz_id)
+
+  def sasl_bind_s(self,dn,mechanism,cred,serverctrls=None,clientctrls=None):
+    """
+    sasl_bind_s(dn, mechanism, cred [,serverctrls=None[,clientctrls=None]]) -> int|str
+    """
+    return self._ldap_call(self._l.sasl_bind_s,dn,mechanism,cred,RequestControlTuples(serverctrls),RequestControlTuples(clientctrls))
+
   def compare_ext(self,dn,attr,value,serverctrls=None,clientctrls=None):
     """
     compare_ext(dn, attr, value [,serverctrls=None[,clientctrls=None]]) -> int
@@ -360,8 +404,7 @@ class SimpleLDAPObject:
     return self.modify_ext(dn,modlist,None,None)
 
   def modify_s(self,dn,modlist):
-    msgid = self.modify(dn,modlist)
-    return self.result(msgid,all=1,timeout=self.timeout)
+    return self.modify_ext_s(dn,modlist,None,None)
 
   def modrdn(self,dn,newrdn,delold=1):
     """
@@ -585,13 +628,25 @@ class SimpleLDAPObject:
         The unbind and unbind_s methods are identical, and are
         synchronous in nature
     """
-    return self._ldap_call(self._l.unbind_ext,RequestControlTuples(serverctrls),RequestControlTuples(clientctrls))
+    res = self._ldap_call(self._l.unbind_ext,RequestControlTuples(serverctrls),RequestControlTuples(clientctrls))
+    try:
+      del self._l
+    except AttributeError:
+      pass
+    return res
 
   def unbind_ext_s(self,serverctrls=None,clientctrls=None):
     msgid = self.unbind_ext(serverctrls,clientctrls)
     if msgid!=None:
-      resp_type, resp_data, resp_msgid, resp_ctrls = self.result3(msgid,all=1,timeout=self.timeout)
-      return resp_type, resp_data, resp_msgid, resp_ctrls
+      result = self.result3(msgid,all=1,timeout=self.timeout)
+    else:
+      result = None
+    if __debug__ and self._trace_level>=1:
+      try:
+        self._trace_file.flush()
+      except AttributeError:
+        pass
+    return result
 
   def unbind(self):
     return self.unbind_ext(None,None)
@@ -648,7 +703,7 @@ class SimpleLDAPObject:
   def read_s(self,dn,filterstr=None,attrlist=None,serverctrls=None,clientctrls=None,timeout=-1):
     """
     Reads and returns a single entry specified by `dn'.
-    
+
     Other attributes just like those passed to `search_ext_s()'
     """
     r = self.search_ext_s(
@@ -663,7 +718,7 @@ class SimpleLDAPObject:
     if r:
       return r[0][1]
     else:
-      raise ldap.NO_SUCH_OBJECT('Empty search result reading entry %s' % (repr(dn)))
+      return None
 
   def read_subschemasubentry_s(self,subschemasubentry_dn,attrs=None):
     """
@@ -699,6 +754,26 @@ class SimpleLDAPObject:
       raise NO_UNIQUE_ENTRY('No or non-unique search result for %s' % (repr(filterstr)))
     return r[0]
 
+  def read_rootdse_s(self, filterstr='(objectClass=*)', attrlist=None):
+    """
+    convenience wrapper around read_s() for reading rootDSE
+    """
+    ldap_rootdse = self.read_s(
+      '',
+      filterstr=filterstr,
+      attrlist=attrlist or ['*', '+'],
+    )
+    return ldap_rootdse  # read_rootdse_s()
+
+  def get_naming_contexts(self):
+    """
+    returns all attribute values of namingContexts in rootDSE
+    if namingContexts is not present (not readable) then empty list is returned
+    """
+    return self.read_rootdse_s(
+      attrlist=['namingContexts']
+    ).get('namingContexts', [])
+
 
 class NonblockingLDAPObject(SimpleLDAPObject):
 
@@ -810,6 +885,9 @@ class ReconnectLDAPObject(SimpleLDAPObject):
     for k,v in self._options:
       SimpleLDAPObject.set_option(self,k,v)
 
+  def passwd_s(self,*args,**kwargs):
+    return self._apply_method_s(SimpleLDAPObject.passwd_s,*args,**kwargs)
+
   def reconnect(self,uri,retry_max=1,retry_delay=60.0):
     # Drop and clean up old connection completely
     # Reconnect
@@ -843,7 +921,6 @@ class ReconnectLDAPObject(SimpleLDAPObject):
             self._trace_file.write('=> delay %s...\n' % (retry_delay))
           time.sleep(retry_delay)
           SimpleLDAPObject.unbind_s(self)
-          del self._l
         else:
           if __debug__ and self._trace_level>=1:
             self._trace_file.write('*** %s reconnect to %s successful => repeat last operation\n' % (
@@ -862,7 +939,6 @@ class ReconnectLDAPObject(SimpleLDAPObject):
       return func(self,*args,**kwargs)
     except ldap.SERVER_DOWN:
       SimpleLDAPObject.unbind_s(self)
-      del self._l
       # Try to reconnect
       self.reconnect(self._uri,retry_max=self._retry_max,retry_delay=self._retry_delay)
       # Re-try last operation
@@ -895,14 +971,19 @@ class ReconnectLDAPObject(SimpleLDAPObject):
     self._store_last_bind(SimpleLDAPObject.sasl_interactive_bind_s,*args,**kwargs)
     return res
 
+  def sasl_bind_s(self,*args,**kwargs):
+    res = self._apply_method_s(SimpleLDAPObject.sasl_bind_s,*args,**kwargs)
+    self._store_last_bind(SimpleLDAPObject.sasl_bind_s,*args,**kwargs)
+    return res
+
   def add_ext_s(self,*args,**kwargs):
     return self._apply_method_s(SimpleLDAPObject.add_ext_s,*args,**kwargs)
 
   def cancel_s(self,*args,**kwargs):
     return self._apply_method_s(SimpleLDAPObject.cancel_s,*args,**kwargs)
 
-  def compare_s(self,*args,**kwargs):
-    return self._apply_method_s(SimpleLDAPObject.compare_s,*args,**kwargs)
+  def compare_ext_s(self,*args,**kwargs):
+    return self._apply_method_s(SimpleLDAPObject.compare_ext_s,*args,**kwargs)
 
   def delete_ext_s(self,*args,**kwargs):
     return self._apply_method_s(SimpleLDAPObject.delete_ext_s,*args,**kwargs)
