@@ -1,10 +1,9 @@
 """
 schema.py - support for subSchemaSubEntry information
-written by Michael Stroeder <michael@stroeder.com>
 
-See http://python-ldap.sourceforge.net for details.
+See http://www.python-ldap.org/ for details.
 
-\$Id: models.py,v 1.28 2005/06/03 10:49:56 stroeder Exp $
+\$Id: models.py,v 1.49 2017/02/18 15:32:01 stroeder Exp $
 """
 
 import UserDict,ldap.cidict
@@ -35,16 +34,28 @@ NOT_HUMAN_READABLE_LDAP_SYNTAXES = {
 class SchemaElement:
   """
   Base class for all schema element classes. Not used directly!
+
+  Arguments:
+
+  schema_element_str
+    String which contains the schema element description to be parsed.
+
+  Class attributes:
+
+  schema_attribute
+    LDAP attribute type containing a certain schema element description
+  token_defaults
+    Dictionary internally used by the schema element parser
+    containing the defaults for certain schema description key-words
   """
   token_defaults = {
     'DESC':(None,),
   }
-  
+
   def __init__(self,schema_element_str=None):
     if schema_element_str:
-      l = split_tokens(schema_element_str,self.token_defaults)
+      l = split_tokens(schema_element_str)
       self.set_id(l[1])
-      assert type(self.get_id())==StringType
       d = extract_tokens(l,self.token_defaults)
       self._set_attrs(l,d)
 
@@ -62,18 +73,18 @@ class SchemaElement:
     assert value is None or type(value)==StringType,TypeError("value has to be of StringType, was %s" % repr(value))
     if value:
       if quoted:
-        return ' %s %s' % (key,repr(value))
+        return " %s '%s'" % (key,value.replace("'","\\'"))
       else:
-        return ' %s %s' % (key,value)
+        return " %s %s" % (key,value)
     else:
-      return ''
+      return ""
 
   def key_list(self,key,values,sep=' ',quoted=0):
     assert type(values)==TupleType,TypeError("values has to be of ListType")
     if not values:
       return ''
     if quoted:
-      quoted_values = map(repr,values)
+      quoted_values = [ "'%s'" % value.replace("'","\\'") for value in values ]
     else:
       quoted_values = values
     if len(values)==1:
@@ -89,17 +100,36 @@ class SchemaElement:
 
 class ObjectClass(SchemaElement):
   """
-  ObjectClassDescription = "(" whsp
-      numericoid whsp      ; ObjectClass identifier
-      [ "NAME" qdescrs ]
-      [ "DESC" qdstring ]
-      [ "OBSOLETE" whsp ]
-      [ "SUP" oids ]       ; Superior ObjectClasses
-      [ ( "ABSTRACT" / "STRUCTURAL" / "AUXILIARY" ) whsp ]
-                           ; default structural
-      [ "MUST" oids ]      ; AttributeTypes
-      [ "MAY" oids ]       ; AttributeTypes
-  whsp ")"
+  Arguments:
+
+  schema_element_str
+    String containing an ObjectClassDescription
+
+  Class attributes:
+
+  oid
+    OID assigned to the object class
+  names
+    This list of strings contains all NAMEs of the object class
+  desc
+    This string contains description text (DESC) of the object class
+  obsolete
+    Integer flag (0 or 1) indicating whether the object class is marked
+    as OBSOLETE in the schema
+  must
+    This list of strings contains NAMEs or OIDs of all attributes
+    an entry of the object class must have
+  may
+    This list of strings contains NAMEs or OIDs of additional attributes
+    an entry of the object class may have
+  kind
+    Kind of an object class:
+    0 = STRUCTURAL,
+    1 = ABSTRACT,
+    2 = AUXILIARY
+  sup
+    This list of strings contains NAMEs or OIDs of object classes
+    this object class is derived from
   """
   schema_attribute = 'objectClasses'
   token_defaults = {
@@ -144,7 +174,7 @@ class ObjectClass(SchemaElement):
     result = [str(self.oid)]
     result.append(self.key_list('NAME',self.names,quoted=1))
     result.append(self.key_attr('DESC',self.desc,quoted=1))
-    result.append(self.key_list('SUP',self.sup))
+    result.append(self.key_list('SUP',self.sup,sep=' $ '))
     result.append({0:'',1:' OBSOLETE'}[self.obsolete])
     result.append({0:' STRUCTURAL',1:' ABSTRACT',2:' AUXILIARY'}[self.kind])
     result.append(self.key_list('MUST',self.must,sep=' $ '))
@@ -153,7 +183,7 @@ class ObjectClass(SchemaElement):
 
 
 AttributeUsage = ldap.cidict.cidict({
-  'userApplication':0,
+  'userApplication':0, # work-around for non-compliant schema
   'userApplications':0,
   'directoryOperation':1,
   'distributedOperation':2,
@@ -163,28 +193,48 @@ AttributeUsage = ldap.cidict.cidict({
 
 class AttributeType(SchemaElement):
   """
-      AttributeTypeDescription = "(" whsp
-            numericoid whsp              ; AttributeType identifier
-          [ "NAME" qdescrs ]             ; name used in AttributeType
-          [ "DESC" qdstring ]            ; description
-          [ "OBSOLETE" whsp ]
-          [ "SUP" woid ]                 ; derived from this other
-                                         ; AttributeType
-          [ "EQUALITY" woid              ; Matching Rule name
-          [ "ORDERING" woid              ; Matching Rule name
-          [ "SUBSTR" woid ]              ; Matching Rule name
-          [ "SYNTAX" whsp noidlen whsp ] ; see section 4.3
-          [ "SINGLE-VALUE" whsp ]        ; default multi-valued
-          [ "COLLECTIVE" whsp ]          ; default not collective
-          [ "NO-USER-MODIFICATION" whsp ]; default user modifiable
-          [ "USAGE" whsp AttributeUsage ]; default userApplications
-          whsp ")"
+  Arguments:
 
-      AttributeUsage =
-          "userApplications"     /
-          "directoryOperation"   /
-          "distributedOperation" / ; DSA-shared
-          "dSAOperation"          ; DSA-specific, value depends on server
+  schema_element_str
+    String containing an AttributeTypeDescription
+
+  Class attributes:
+
+  oid
+    OID assigned to the attribute type
+  names
+    This list of strings contains all NAMEs of the attribute type
+  desc
+    This string contains description text (DESC) of the attribute type
+  obsolete
+    Integer flag (0 or 1) indicating whether the attribute type is marked
+    as OBSOLETE in the schema
+  single_value
+    Integer flag (0 or 1) indicating whether the attribute must
+    have only one value
+  syntax
+    String contains OID of the LDAP syntax assigned to the attribute type
+  no_user_mod
+    Integer flag (0 or 1) indicating whether the attribute is modifiable
+    by a client application
+  equality
+    String contains NAME or OID of the matching rule used for
+    checking whether attribute values are equal
+  substr
+    String contains NAME or OID of the matching rule used for
+    checking whether an attribute value contains another value
+  ordering
+    String contains NAME or OID of the matching rule used for
+    checking whether attribute values are lesser-equal than
+  usage
+    USAGE of an attribute type:
+    0 = userApplications
+    1 = directoryOperation,
+    2 = distributedOperation,
+    3 = dSAOperation
+  sup
+    This list of strings contains NAMEs or OIDs of attribute types
+    this attribute type is derived from
   """
   schema_attribute = 'attributeTypes'
   token_defaults = {
@@ -199,7 +249,9 @@ class AttributeType(SchemaElement):
     'SINGLE-VALUE':None,
     'COLLECTIVE':None,
     'NO-USER-MODIFICATION':None,
-    'USAGE':('userApplications',)
+    'USAGE':('userApplications',),
+    'X-ORIGIN':(None,),
+    'X-ORDERED':(None,),
   }
 
   def _set_attrs(self,l,d):
@@ -210,28 +262,31 @@ class AttributeType(SchemaElement):
     self.equality = d['EQUALITY'][0]
     self.ordering = d['ORDERING'][0]
     self.substr = d['SUBSTR'][0]
-    syntax = d['SYNTAX'][0]
-    if syntax is None:
+    self.x_origin = d['X-ORIGIN'][0]
+    self.x_ordered = d['X-ORDERED'][0]
+    try:
+      syntax = d['SYNTAX'][0]
+    except IndexError:
       self.syntax = None
       self.syntax_len = None
     else:
-      try:
-        self.syntax,syntax_len = d['SYNTAX'][0].split("{")
-      except ValueError:
-        self.syntax = d['SYNTAX'][0]
+      if syntax is None:
+        self.syntax = None
         self.syntax_len = None
-        for i in l:
-          if i.startswith("{") and i.endswith("}"):
-            self.syntax_len=long(i[1:-1])
       else:
-        self.syntax_len = long(syntax_len[:-1])
+        try:
+          self.syntax,syntax_len = d['SYNTAX'][0].split("{")
+        except ValueError:
+          self.syntax = d['SYNTAX'][0]
+          self.syntax_len = None
+          for i in l:
+            if i.startswith("{") and i.endswith("}"):
+              self.syntax_len=long(i[1:-1])
+        else:
+          self.syntax_len = long(syntax_len[:-1])
     self.single_value = d['SINGLE-VALUE']!=None
     self.collective = d['COLLECTIVE']!=None
     self.no_user_mod = d['NO-USER-MODIFICATION']!=None
-    try:
-      self.usage = AttributeUsage[d['USAGE'][0]]
-    except KeyError:
-      raise
     self.usage = AttributeUsage.get(d['USAGE'][0],0)
     assert type(self.names)==TupleType
     assert self.desc is None or type(self.desc)==StringType
@@ -247,7 +302,7 @@ class AttributeType(SchemaElement):
     result = [str(self.oid)]
     result.append(self.key_list('NAME',self.names,quoted=1))
     result.append(self.key_attr('DESC',self.desc,quoted=1))
-    result.append(self.key_list('SUP',self.sup))
+    result.append(self.key_list('SUP',self.sup,sep=' $ '))
     result.append({0:'',1:' OBSOLETE'}[self.obsolete])
     result.append(self.key_attr('EQUALITY',self.equality))
     result.append(self.key_attr('ORDERING',self.ordering))
@@ -266,33 +321,45 @@ class AttributeType(SchemaElement):
         3:" USAGE dSAOperation",
       }[self.usage]
     )
+    result.append(self.key_attr('X-ORIGIN',self.x_origin,quoted=1))
+    result.append(self.key_attr('X-ORDERED',self.x_ordered,quoted=1))
     return '( %s )' % ''.join(result)
 
 
 class LDAPSyntax(SchemaElement):
   """
-  SyntaxDescription = "(" whsp
-      numericoid whsp
-      [ "DESC" qdstring ]
-      whsp ")"
+  SyntaxDescription
+
+  oid
+    OID assigned to the LDAP syntax
+  desc
+    This string contains description text (DESC) of the LDAP syntax
+  not_human_readable
+    Integer flag (0 or 1) indicating whether the attribute type is marked
+    as not human-readable (X-NOT-HUMAN-READABLE)
   """
   schema_attribute = 'ldapSyntaxes'
   token_defaults = {
     'DESC':(None,),
     'X-NOT-HUMAN-READABLE':(None,),
+    'X-BINARY-TRANSFER-REQUIRED':(None,),
+    'X-SUBST':(None,),
   }
 
   def _set_attrs(self,l,d):
     self.desc = d['DESC'][0]
+    self.x_subst = d['X-SUBST'][0]
     self.not_human_readable = \
       NOT_HUMAN_READABLE_LDAP_SYNTAXES.has_key(self.oid) or \
       d['X-NOT-HUMAN-READABLE'][0]=='TRUE'
+    self.x_binary_transfer_required = d['X-BINARY-TRANSFER-REQUIRED'][0]=='TRUE'
     assert self.desc is None or type(self.desc)==StringType
     return
-                                  
+
   def __str__(self):
     result = [str(self.oid)]
     result.append(self.key_attr('DESC',self.desc,quoted=1))
+    result.append(self.key_attr('X-SUBST',self.x_subst,quoted=1))
     result.append(
       {0:'',1:" X-NOT-HUMAN-READABLE 'TRUE'"}[self.not_human_readable]
     )
@@ -301,13 +368,24 @@ class LDAPSyntax(SchemaElement):
 
 class MatchingRule(SchemaElement):
   """
-  MatchingRuleDescription = "(" whsp
-      numericoid whsp  ; MatchingRule identifier
-      [ "NAME" qdescrs ]
-      [ "DESC" qdstring ]
-      [ "OBSOLETE" whsp ]
-      "SYNTAX" numericoid
-  whsp ")"
+  Arguments:
+
+  schema_element_str
+    String containing an MatchingRuleDescription
+
+  Class attributes:
+
+  oid
+    OID assigned to the matching rule
+  names
+    This list of strings contains all NAMEs of the matching rule
+  desc
+    This string contains description text (DESC) of the matching rule
+  obsolete
+    Integer flag (0 or 1) indicating whether the matching rule is marked
+    as OBSOLETE in the schema
+  syntax
+    String contains OID of the LDAP syntax this matching rule is usable with
   """
   schema_attribute = 'matchingRules'
   token_defaults = {
@@ -339,14 +417,25 @@ class MatchingRule(SchemaElement):
 
 class MatchingRuleUse(SchemaElement):
   """
-  MatchingRuleUseDescription = "(" whsp
-     numericoid 
-     [ space "NAME" space qdescrs ]
-     [ space "DESC" space qdstring ]
-     [ space "OBSOLETE" ]
-     space "APPLIES" space oids    ;  AttributeType identifiers
-     extensions
-     whsp ")" 
+  Arguments:
+
+  schema_element_str
+    String containing an MatchingRuleUseDescription
+
+  Class attributes:
+
+  oid
+    OID of the accompanying matching rule
+  names
+    This list of strings contains all NAMEs of the matching rule
+  desc
+    This string contains description text (DESC) of the matching rule
+  obsolete
+    Integer flag (0 or 1) indicating whether the matching rule is marked
+    as OBSOLETE in the schema
+  applies
+    This list of strings contains NAMEs or OIDs of attribute types
+    for which this matching rule is used
   """
   schema_attribute = 'matchingRuleUse'
   token_defaults = {
@@ -378,16 +467,36 @@ class MatchingRuleUse(SchemaElement):
 
 class DITContentRule(SchemaElement):
   """
-  DITContentRuleDescription = LPAREN WSP
-      numericoid                 ; object identifer
-      [ SP "NAME" SP qdescrs ]   ; short names
-      [ SP "DESC" SP qdstring ]  ; description
-      [ SP "OBSOLETE" ]          ; not active
-      [ SP "AUX" SP oids ]       ; auxiliary object classes
-      [ SP "MUST" SP oids ]      ; attribute types
-      [ SP "MAY" SP oids ]       ; attribute types
-      [ SP "NOT" SP oids ]       ; attribute types
-      extensions WSP RPAREN      ; extensions
+  Arguments:
+
+  schema_element_str
+    String containing an DITContentRuleDescription
+
+  Class attributes:
+
+  oid
+    OID of the accompanying structural object class
+  names
+    This list of strings contains all NAMEs of the DIT content rule
+  desc
+    This string contains description text (DESC) of the DIT content rule
+  obsolete
+    Integer flag (0 or 1) indicating whether the DIT content rule is marked
+    as OBSOLETE in the schema
+  aux
+    This list of strings contains NAMEs or OIDs of all auxiliary
+    object classes usable in an entry of the object class
+  must
+    This list of strings contains NAMEs or OIDs of all attributes
+    an entry of the object class must have which may extend the
+    list of required attributes of the object classes of an entry
+  may
+    This list of strings contains NAMEs or OIDs of additional attributes
+    an entry of the object class may have which may extend the
+    list of optional attributes of the object classes of an entry
+  nots
+    This list of strings contains NAMEs or OIDs of attributes which
+    may not be present in an entry of the object class
   """
   schema_attribute = 'dITContentRules'
   token_defaults = {
@@ -431,14 +540,27 @@ class DITContentRule(SchemaElement):
 
 class DITStructureRule(SchemaElement):
   """
-  DITStructureRuleDescription = LPAREN WSP
-      ruleid                     ; rule identifier
-      [ SP "NAME" SP qdescrs ]   ; short names
-      [ SP "DESC" SP qdstring ]  ; description
-      [ SP "OBSOLETE" ]          ; not active
-      SP "FORM" SP oid           ; NameForm
-      [ SP "SUP" ruleids ]       ; superior rules
-      extensions WSP RPAREN      ; extensions
+  Arguments:
+
+  schema_element_str
+    String containing an DITStructureRuleDescription
+
+  Class attributes:
+
+  ruleid
+    rule ID of the DIT structure rule (only locally unique)
+  names
+    This list of strings contains all NAMEs of the DIT structure rule
+  desc
+    This string contains description text (DESC) of the DIT structure rule
+  obsolete
+    Integer flag (0 or 1) indicating whether the DIT content rule is marked
+    as OBSOLETE in the schema
+  form
+    List of strings with NAMEs or OIDs of associated name forms
+  sup
+    List of strings with NAMEs or OIDs of allowed structural object classes
+    of superior entries in the DIT
   """
   schema_attribute = 'dITStructureRules'
 
@@ -481,22 +603,40 @@ class DITStructureRule(SchemaElement):
 
 class NameForm(SchemaElement):
   """
-  NameFormDescription = LPAREN WSP
-      numericoid                 ; object identifer
-      [ SP "NAME" SP qdescrs ]   ; short names
-      [ SP "DESC" SP qdstring ]  ; description
-      [ SP "OBSOLETE" ]          ; not active
-      SP "OC" SP oid             ; structural object class
-      SP "MUST" SP oids          ; attribute types
-      [ SP "MAY" SP oids ]       ; attribute types
-      extensions WSP RPAREN      ; extensions
+  Arguments:
+
+  schema_element_str
+    String containing an NameFormDescription
+
+  Class attributes:
+
+  oid
+    OID of the name form
+  names
+    This list of strings contains all NAMEs of the name form
+  desc
+    This string contains description text (DESC) of the name form
+  obsolete
+    Integer flag (0 or 1) indicating whether the name form is marked
+    as OBSOLETE in the schema
+  form
+    List of strings with NAMEs or OIDs of associated name forms
+  oc
+    String with NAME or OID of structural object classes this name form
+    is usable with
+  must
+    This list of strings contains NAMEs or OIDs of all attributes
+    an RDN must contain
+  may
+    This list of strings contains NAMEs or OIDs of additional attributes
+    an RDN may contain
   """
   schema_attribute = 'nameForms'
   token_defaults = {
     'NAME':(()),
     'DESC':(None,),
     'OBSOLETE':None,
-    'OC':(()),
+    'OC':(None,),
     'MUST':(()),
     'MAY':(()),
   }
@@ -505,13 +645,13 @@ class NameForm(SchemaElement):
     self.names = d['NAME']
     self.desc = d['DESC'][0]
     self.obsolete = d['OBSOLETE']!=None
-    self.oc = d['OC']
+    self.oc = d['OC'][0]
     self.must = d['MUST']
     self.may = d['MAY']
     assert type(self.names)==TupleType
     assert self.desc is None or type(self.desc)==StringType
     assert type(self.obsolete)==BooleanType and (self.obsolete==0 or self.obsolete==1)
-    assert type(self.oc)==TupleType
+    assert type(self.oc)==StringType
     assert type(self.must)==TupleType
     assert type(self.may)==TupleType
     return
@@ -521,7 +661,7 @@ class NameForm(SchemaElement):
     result.append(self.key_list('NAME',self.names,quoted=1))
     result.append(self.key_attr('DESC',self.desc,quoted=1))
     result.append({0:'',1:' OBSOLETE'}[self.obsolete])
-    result.append(self.key_list('OC',self.oc,sep=' $ '))
+    result.append(self.key_attr('OC',self.oc))
     result.append(self.key_list('MUST',self.must,sep=' $ '))
     result.append(self.key_list('MAY',self.may,sep=' $ '))
     return '( %s )' % ''.join(result)
@@ -530,7 +670,7 @@ class NameForm(SchemaElement):
 class Entry(UserDict.UserDict):
   """
   Schema-aware implementation of an LDAP entry class.
-  
+
   Mainly it holds the attributes in a string-keyed dictionary with
   the OID as key.
   """
@@ -564,6 +704,9 @@ class Entry(UserDict.UserDict):
     for key in dict.keys():
       self[key] = dict[key]
 
+  def __contains__(self,key):
+    return self.has_key(key)
+
   def __getitem__(self,nameoroid):
     return self.data[self._at2key(nameoroid)]
 
