@@ -1,10 +1,9 @@
 """
-ldapurl - handling of LDAP URLs as described in RFC 2255
-written by Michael Stroeder <michael@stroeder.com>
+ldapurl - handling of LDAP URLs as described in RFC 4516
 
-See http://python-ldap.sourceforge.net for details.
+See http://www.python-ldap.org/ for details.
 
-\$Id: ldapurl.py,v 1.33 2005/06/01 22:11:20 stroeder Exp $
+\$Id: ldapurl.py,v 1.94 2017/07/12 17:30:12 stroeder Exp $
 
 Python compability note:
 This module only works with Python 2.0+ since
@@ -12,7 +11,7 @@ This module only works with Python 2.0+ since
 2. list comprehensions are used.
 """
 
-__version__ = '0.5.2'
+__version__ = '2.4.41'
 
 __all__ = [
   # constants
@@ -31,15 +30,24 @@ from urllib import quote,unquote
 LDAP_SCOPE_BASE = 0
 LDAP_SCOPE_ONELEVEL = 1
 LDAP_SCOPE_SUBTREE = 2
+LDAP_SCOPE_SUBORDINATES = 3
 
-SEARCH_SCOPE_STR = {None:'',0:'base',1:'one',2:'sub'}
+SEARCH_SCOPE_STR = {
+  None:'',
+  LDAP_SCOPE_BASE:'base',
+  LDAP_SCOPE_ONELEVEL:'one',
+  LDAP_SCOPE_SUBTREE:'sub',
+  LDAP_SCOPE_SUBORDINATES:'subordinates',
+}
 
 SEARCH_SCOPE = {
   '':None,
-  # the search scope strings defined in RFC2255
+  # the search scope strings defined in RFC 4516
   'base':LDAP_SCOPE_BASE,
   'one':LDAP_SCOPE_ONELEVEL,
   'sub':LDAP_SCOPE_SUBTREE,
+  # from draft-sermersheim-ldap-subordinate-scope
+  'subordinates':LDAP_SCOPE_SUBORDINATES,
 }
 
 # Some widely used types
@@ -66,24 +74,15 @@ def ldapUrlEscape(s):
 class LDAPUrlExtension:
   """
   Class for parsing and unparsing LDAP URL extensions
-  as described in RFC 2255.
-
-  BNF definition of LDAP URL extensions:
-
-       extensions = extension *("," extension)
-       extension  = ["!"] extype ["=" exvalue]
-       extype     = token / xtoken
-       exvalue    = LDAPString from section 4.1.2 of [2]
-       token      = oid from section 4.1 of [3]
-       xtoken     = ("X-" / "x-") token
+  as described in RFC 4516.
 
   Usable class attributes:
-  critical
-        Boolean integer marking the extension as critical
-  extype    
-        Type of extension
-  exvalue
-        Value of extension
+    critical
+          Boolean integer marking the extension as critical
+    extype
+          Type of extension
+    exvalue
+          Value of extension
   """
 
   def __init__(self,extensionStr=None,critical=0,extype=None,exvalue=None):
@@ -102,16 +101,24 @@ class LDAPUrlExtension:
     self.critical = extension[0]=='!'
     if extension[0]=='!':
       extension = extension[1:].strip()
-    self.extype,self.exvalue = extension.split('=',1)
+    try:
+      self.extype,self.exvalue = extension.split('=',1)
+    except ValueError:
+      # No value, just the extype
+      self.extype,self.exvalue = extension,None
+    else:
+      self.exvalue = unquote(self.exvalue.strip())
     self.extype = self.extype.strip()
-    self.exvalue = unquote(self.exvalue.strip())
 
   def unparse(self):
-    return '%s%s=%s' % (
-      '!'*(self.critical>0),
-      self.extype,(self.exvalue or '').replace(',',r'%2C')
-    )
-    
+    if self.exvalue is None:
+      return '%s%s' % ('!'*(self.critical>0),self.extype)
+    else:
+      return '%s%s=%s' % (
+        '!'*(self.critical>0),
+        self.extype,quote(self.exvalue or '')
+      )
+
   def __str__(self):
     return self.unparse()
 
@@ -176,11 +183,12 @@ class LDAPUrlExtensions(UserDict.UserDict):
       "other has to be instance of %s" % (self.__class__)
     )
     return self.data==other.data
-    
+
   def parse(self,extListStr):
     for extension_str in extListStr.strip().split(','):
-      e = LDAPUrlExtension(extension_str)
-      self[e.extype] = e
+      if extension_str:
+        e = LDAPUrlExtension(extension_str)
+        self[e.extype] = e
 
   def unparse(self):
     return ','.join([ v.unparse() for v in self.values() ])
@@ -189,18 +197,7 @@ class LDAPUrlExtensions(UserDict.UserDict):
 class LDAPUrl:
   """
   Class for parsing and unparsing LDAP URLs
-  as described in RFC 2255.
-
-  BNF definition of LDAP URL:
-
-    hostport     host:port
-    dn           distinguished name
-    attributes   list with attributes
-    scope        search scope string
-    filter       LDAP search filter
-    ldapurl    = scheme "://" [hostport] ["/"
-                     [dn ["?" [attrs] ["?" [scope]
-                     ["?" [filter] ["?" extensions]]]]]]
+  as described in RFC 4516.
 
   Usable class attributes:
     urlscheme
@@ -215,7 +212,7 @@ class LDAPUrl:
         integer search scope for ldap-module
     filterstr
         String representation of LDAP Search Filters
-        (see RFC 2254)
+        (see RFC 4515)
     extensions
         Dictionary used as extensions store
     who
@@ -267,11 +264,11 @@ class LDAPUrl:
     urlscheme,host,dn,attrs,scope,filterstr,extensions
     """
     if not isLDAPUrl(ldap_url):
-      raise ValueError,'Parameter ldap_url does not seem to be a LDAP URL.'
+      raise ValueError('Value %s for ldap_url does not seem to be a LDAP URL.' % (repr(ldap_url)))
     scheme,rest = ldap_url.split('://',1)
     self.urlscheme = scheme.strip()
     if not self.urlscheme in ['ldap','ldaps','ldapi']:
-      raise ValueError,'LDAP URL contains unsupported URL scheme %s.' % (self.urlscheme)
+      raise ValueError('LDAP URL contains unsupported URL scheme %s.' % (self.urlscheme))
     slash_pos = rest.find('/')
     qemark_pos = rest.find('?')
     if (slash_pos==-1) and (qemark_pos==-1):
@@ -291,8 +288,8 @@ class LDAPUrl:
         # Do not eat question mark
         rest = rest[qemark_pos:]
       else:
-        raise ValueError,'Something completely weird happened!'
-    paramlist=rest.split('?')
+        raise ValueError('Something completely weird happened!')
+    paramlist=rest.split('?',4)
     paramlist_len = len(paramlist)
     if paramlist_len>=1:
       self.dn = unquote(paramlist[0]).strip()
@@ -303,7 +300,7 @@ class LDAPUrl:
       try:
         self.scope = SEARCH_SCOPE[scope]
       except KeyError:
-        raise ValueError,"Search scope must be either one of base, one or sub. LDAP URL contained %s" % (repr(scope))
+        raise ValueError('Invalid search scope %s' % (repr(scope)))
     if paramlist_len>=4:
       filterstr = paramlist[3].strip()
       if not filterstr:
@@ -311,8 +308,11 @@ class LDAPUrl:
       else:
         self.filterstr = unquote(filterstr)
     if paramlist_len>=5:
-      self.extensions = LDAPUrlExtensions()
-      self.extensions.parse(paramlist[4])
+      if paramlist[4]:
+        self.extensions = LDAPUrlExtensions()
+        self.extensions.parse(paramlist[4])
+      else:
+        self.extensions = None
     return
 
   def applyDefaults(self,defaults):
@@ -364,9 +364,18 @@ class LDAPUrl:
     if self.extensions:
       ldap_url = ldap_url+'?'+self.extensions.unparse()
     return ldap_url
-  
+
   def htmlHREF(self,urlPrefix='',hrefText=None,hrefTarget=None):
-    """Complete """
+    """
+    Returns a string with HTML link for this LDAP URL.
+
+    urlPrefix
+        Prefix before LDAP URL (e.g. for addressing another web-based client)
+    hrefText
+        link text/description
+    hrefTarget
+        string added as link target attribute
+    """
     assert type(urlPrefix)==StringType, "urlPrefix must be StringType"
     if hrefText is None:
       hrefText = self.unparse()
@@ -394,14 +403,16 @@ class LDAPUrl:
   def __getattr__(self,name):
     if self.attr2extype.has_key(name):
       extype = self.attr2extype[name]
-      if self.extensions.has_key(extype):
+      if self.extensions and \
+         self.extensions.has_key(extype) and \
+         not self.extensions[extype].exvalue is None:
         result = unquote(self.extensions[extype].exvalue)
       else:
         return None
     else:
-      raise AttributeError,"%s has no attribute %s" % (
+      raise AttributeError('%s has no attribute %s' % (
         self.__class__.__name__,name
-      )
+      ))
     return result # __getattr__()
 
   def __setattr__(self,name,value):
@@ -421,10 +432,11 @@ class LDAPUrl:
   def __delattr__(self,name):
     if self.attr2extype.has_key(name):
       extype = self.attr2extype[name]
-      try:
-        del self.extensions[extype]
-      except KeyError:
-        pass
+      if self.extensions:
+        try:
+          del self.extensions[extype]
+        except KeyError:
+          pass
     else:
       del self.__dict__[name]
 
